"""
This package holds the common libraries used by all Python parts of the Optimal BPM system.
It includes messaging, queue management, database initialization, decorators, and logging
"""
__author__ = 'Nicklas Borjesson'
